#ifndef FERI_ATTRACTOR
#define FERI_ATTRACTOR
// -----------------------------------------------
// ATTRACTOR ANALYSIS WITH GRAPH REPRESENTATION
// RICKER MAP
//
// Main Program
// Ferenc Bartha
//
// using libraries:
// - CAPD
// - BOOST-GRAPH

// -----------------------------------------------

// GRAPHICS
//#define FERI_GRAPHICS

// GRAPHICS : DRAW EXTRA INFORMATION
//#define FERI_GRAPHICS_DRAW_EXTRA

// -----------------------------------------------

#include "main.hpp"
#include "typedefs.hpp"
#include "routines.hpp"
#include "vectorfield.hpp"

// -----------------------------------------------

// vectorfield
typedef VectorField<interval, FERI_DIMENSION> VectorFieldType;

// -----------------------------------------------

// trying to prove for a given parameter range
void simulateRicker(interval parama, double initialBoxSize)
{
	boost::timer rickerTime;
	VectorType V, enclosure, attDomain;
	GraphType GRAPH(0);
	double boxSize, runTime;
	int iter, max_iter;
	bool stop;



	// max number of iterations, 0 = unlimited
	max_iter = 100;



	#ifdef FERI_CHECK_STRIPE
		bool leftSeparated, rightEmpty;
		const double leftStripe = 0.11;									// suspected separating stripe

		leftSeparated = false;
		rightEmpty = false;
	#endif



	VectorFieldType TheMap(parama);



	#ifdef FERI_GRAPHICS
		openGraphics(500,500,WHITE,BLACK);
		double SCREENSIZE = 8;
		#ifdef FERI_ROTATED
			rootFrame.setWorldCoord(0., - SCREENSIZE / 2.0, SCREENSIZE, SCREENSIZE / 2.0);
		#else
			rootFrame.setWorldCoord(0., 0., SCREENSIZE, SCREENSIZE);
		#endif

		TheMap.resetGraphics(SCREENSIZE);
	#endif



	// output
	#ifdef FERI_OUTPUT_PER_PARAM
		std::string output = outputBase + "_" + boost::lexical_cast<std::string>(parama.leftBound()) + "_" + boost::lexical_cast<std::string>(parama.rightBound()) + ".txt";

		std::ofstream ostream(output.c_str());
		ostream.precision(oPrecision);

		ostream << "Simulation for Ricker Map, with parameter = " << parama << std::endl;
	#endif



	// we find the invariant set
	// we have finite convergence and stopping tolerance
	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "Constructing the invariant set... " << std::flush;
	#endif

	TheMap.constructInvariantSet(V);

	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "done" << std::endl;
		ostream << "Starting simulation on the invariant set X x Y: " << std::endl;
		ostream << "X = " << V(1) << std::endl;
		ostream << "Y = " << V(2) << std::endl;
	#endif



	// finding the local domain of attraction
	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "Constructing the local attraction domain... " << std::flush;
	#endif

	#ifdef FERI_LYAPUNOV
		TheMap.findLyapAttractionDomain(V, attDomain);
	#else
		TheMap.findAttractionDomain(attDomain);
	#endif

	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "done" << std::endl;
		ostream << "Local attraction domain is: " << attDomain << std::endl;
	#endif



	// building the initial graph
	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "Filling the initial graph with boxes... " << std::flush;
	#endif

	#ifdef FERI_ROTATED
		TheMap.buildRotatedInitialBoxes(V, initialBoxSize, GRAPH);
	#else
		buildInitialBoxes(V, initialBoxSize, GRAPH);
	#endif

	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "done" << std::endl;
	#endif



	// adding edges
	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "Adding edges to the graph... " << std::flush;
	#endif

	addInitialEdges(TheMap, GRAPH);

	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "done" << std::endl;
		reportSizeOfGraph(GRAPH, ostream);



		// evaluating enclosure for the invariant set
		evaluateEnclosure(boxSize, GRAPH, enclosure);
		ostream << "  ---- Enclosure is: " << enclosure << " with box size: " << boxSize << std::endl;
	#endif



	// MAIN LOOP STARTS
	iter = 0;
	do {
		++iter;

		#ifdef FERI_GRAPHICS
			TheMap.resetGraphics(SCREENSIZE);
		#endif



		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << " -------------- Iteration no. " << iter << std::endl;


			#ifdef FERI_SOURCESANDSINKS
				ostream << "Removing boxes that are sources or sinks... " << std::flush;
			#else
				ostream << "Keeping only SCCs... " << std::flush;
			#endif
		#endif



		#ifdef FERI_SOURCESANDSINKS
			removeSourcesAndSinks(GRAPH, 5);
		#else
			checkStronglyConnectedComponents(GRAPH);
		#endif



		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "done" << std::endl;
			reportSizeOfGraph(GRAPH, ostream);
		#endif



		// bisecting remaining boxes
		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "Bisecting boxes... " << std::flush;
		#endif

		bisectBoxes(GRAPH);

		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "done" << std::endl;
			reportSizeOfGraph(GRAPH, ostream);
		#endif



		// checking the mapping between new boxes
		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "Keeping only the good edges... " << std::flush;
		#endif

		keepOnlyGoodEdges(TheMap, GRAPH);

		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "done" << std::endl;
			reportSizeOfGraph(GRAPH, ostream);



			// evaluating enclosure for the invariant set
			if (evaluateEnclosure(boxSize, GRAPH, enclosure)) {
				ostream << "  ---- Enclosure is: " << enclosure << " with box size: " << boxSize << std::endl;
			}
		#endif



		#ifdef FERI_GRAPHICS

			if (evaluateEnclosure(boxSize, GRAPH, enclosure)) {

				SCREENSIZE = capd::max(capd::abs(enclosure(1)).rightBound(), capd::abs(enclosure(2)).rightBound());

				#ifdef FERI_ROTATED
					rootFrame.setWorldCoord(0., - SCREENSIZE / 2.0, SCREENSIZE, SCREENSIZE / 2.0);
				#else
					rootFrame.setWorldCoord(0., 0., SCREENSIZE, SCREENSIZE);
				#endif

				plotBoxes(GRAPH, 1, 2);

				#ifdef FERI_GRAPHICS_DRAW_EXTRA
					TheMap.plotFixedPoint(SCREENSIZE);
				#endif
			}

		#endif



		#ifdef FERI_CHECK_STRIPE

			#ifdef FERI_OUTPUT_PER_PARAM
				ostream << "Checking the separating stripe... " << std::flush;
			#endif

			// checking the separating stripe
			TheMap.checkStripe(leftStripe, GRAPH, leftSeparated, rightEmpty);

			#ifdef FERI_OUTPUT_PER_PARAM
				ostream << "done" << std::endl;
				ostream << "left side separated: " << leftSeparated << ", right side has no connections to left: " << rightEmpty << std::endl;
			#endif

		#endif



		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "Checking the attracted boxes... " << std::flush;
		#endif



		// checking if all the boxes are attracted
		#ifdef FERI_ROTATED
			stop = TheMap.checkRotatedAttractionDomain(attDomain, GRAPH);
		#else
			stop = checkAttractionDomain(attDomain, GRAPH);
		#endif



		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "done" << std::endl;
		#endif



	} while( not(stop) && ((iter < max_iter) || (max_iter == 0)) );
	// MAIN LOOP DONE



	runTime = rickerTime.elapsed();



	// reached max number of iterations
	if ( not(stop) ) {

		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "Running Time : " << runTime << std::endl;
			ostream << "Could not validate Fixed Point" << std::endl;

			ostream.close();
		#endif

		std::ofstream ostream(errorFile, std::ios::app);

		ostream.precision(oPrecision);

		if (evaluateEnclosure(boxSize, GRAPH, enclosure)) {
			ostream.width(0);
			ostream << "[";
			ostream.width(oPrecision + 2);
			ostream << parama.leftBound();
			ostream.width(0);
			ostream << ",";
			ostream.width(oPrecision + 2);
			ostream << parama.rightBound();
			ostream.width(0);
			ostream << "]";
			ostream.width(oPrecision + 8);
			ostream << runTime;
			ostream.width(19);
			ostream << iter;
			ostream.width(28);
			ostream << boxSize;
			ostream.width(0);
			ostream << "      " << enclosure << std::endl;
		} else {
			ostream.width(0);
			ostream << "[";
			ostream.width(oPrecision + 2);
			ostream << parama.leftBound();
			ostream.width(0);
			ostream << ",";
			ostream.width(oPrecision + 2);
			ostream << parama.rightBound();
			ostream.width(0);
			ostream << "]";
			ostream.width(oPrecision + 8);
			ostream << runTime;
			ostream.width(19);
			ostream << iter << std::endl;
		}

		ostream << std::flush;

		ostream.close();



	// fixed point is validated
	} else {

		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "Running Time : " << runTime << std::endl;
			ostream << "Attracting Fixed Point validated" << std::endl;

			ostream.close();
		#endif

		std::ofstream ostream(successFile, std::ios::app);

		ostream.precision(oPrecision);

		if (evaluateEnclosure(boxSize, GRAPH, enclosure)) {
			ostream.width(0);
			ostream << "[";
			ostream.width(oPrecision + 2);
			ostream << parama.leftBound();
			ostream.width(0);
			ostream << ",";
			ostream.width(oPrecision + 2);
			ostream << parama.rightBound();
			ostream.width(0);
			ostream << "]";
			ostream.width(oPrecision + 8);
			ostream << runTime;
			ostream.width(19);
			ostream << iter;
			ostream.width(28);
			ostream << boxSize;
			ostream.width(0);
			ostream << "      " << enclosure << std::endl;
		} else {
			ostream.width(0);
			ostream << "[";
			ostream.width(oPrecision + 2);
			ostream << parama.leftBound();
			ostream.width(0);
			ostream << ",";
			ostream.width(oPrecision + 2);
			ostream << parama.rightBound();
			ostream.width(0);
			ostream << "]";
			ostream.width(oPrecision + 8);
			ostream << runTime;
			ostream.width(19);
			ostream << iter << std::endl;
		}

		ostream << std::flush;

		ostream.close();

	}

}



// -----------------------------------------------



int main(int argc, char* argv[])
{

	interval paramRange = interval(0.9,0.9991);							// the parameter range
	double paramsize = 1e-8;											// size of parameter segments to be analyzed
	double initialBoxSize = 1e-1;										// initial box size



	// we have overriding command line parameters
	if (argc > 1) {
		paramRange = interval(atof(argv[1]), atof(argv[2]));
		paramsize = atof(argv[3]);
		initialBoxSize = atof(argv[4]);
		#ifdef FERI_OPENMP
			omp_set_num_threads(atoi(argv[5]));
		#endif
	}



	// printing to succesfile
	std::ofstream ostream(successFile, std::ios::app);
	ostream << std::setw(2*oPrecision+7) << "Parameter range" << "      " << std::setw(oPrecision+2) << "Run time" << "      " << std::setw(13) << "Iteration no." << std::setw(28) << "box size" << std::setw(0) << "       enclosure" << std::endl;
	ostream.close();



	// printing to errorfile
	ostream.open(errorFile, std::ios::app);
	ostream << std::setw(2*oPrecision+7) << "Parameter range" << "      " << std::setw(oPrecision+2) << "Run time" << "      " << std::setw(13) << "Iteration no." << std::setw(28) << "box size" << std::setw(0) << "       enclosure" << std::endl;
	ostream.close();



	// choice of parameter from range
	long int iter, size;
	interval a = paramRange.left() + interval(0, paramsize);
	size = (long int) (diam(paramRange).rightBound() / paramsize) + 1;
	std::vector<interval> params;
	params.reserve(size);

	while (a.rightBound() < paramRange.rightBound()) {
		params.push_back(a);
		a = interval(a.rightBound(), a.rightBound() + paramsize);
	}
	a = interval(a.leftBound(), paramRange.rightBound());
	params.push_back(a);

	std::cout.precision(oPrecision);
	std::cout << "Parameter range " << paramRange << " is divided into " << params.size() << " pieces. (paramSize: "<< paramsize << ", initialBoxSize: " << initialBoxSize << ")" << std::endl;



	// checking every parameter chunk
	#ifdef FERI_OPENMP
		#pragma omp parallel shared(params, initialBoxSize)
	#endif
	{
		#ifdef FERI_OPENMP
			#pragma omp for schedule(dynamic)
		#endif
		for(iter = 0; iter < (long int) params.size(); ++iter)
		{
			simulateRicker(params[iter], initialBoxSize);
		}
	}

}

// ~ end of file ~
#endif
